import functools
import multiprocessing
from contextlib import contextmanager
from pathlib import Path
from typing import Any, List, Optional, Union

import dask
from dask.delayed import Delayed

from dutil.pipeline._cached import CachedResultItem, _kw_is_private, cached


class DelayedParameter:
    """Delayed parameter = a Delayed object that can change the returned value

    Important! Methods `update` and `context` do not work with dask distributed.

    :param name: parameter name
    :param value: parameter value
    """

    def __init__(self, name, value=None):
        self._name = name
        self._value = value
        self._delayed = dask.delayed(name=name)(lambda: self._value)()
        self._lock_context = multiprocessing.Lock()

    def set(self, value) -> None:
        """Permanently change the value of this parameter"""
        self._value = value

    def __call__(self) -> Delayed:
        """Get a Delayed object"""
        return self._delayed

    @contextmanager
    def context(self, value):
        """Change the value of this parameter within a context"""
        with self._lock_context:
            old_value = self._value
            self.set(value)
            yield
            self.set(old_value)


class DelayedParameters:
    """A dictionary of delayed parameters

    Important! Methods `update` and `context` do not work with dask distributed.
    """

    def __init__(self):
        self._params = {}
        self._param_delayed = {}
        self._lock_context = multiprocessing.Lock()

    def create(self, name: str, value: Any = None) -> Delayed:
        """Create a new parameter and return a delayed object"""
        if name in self._params:
            raise KeyError(f"Parameter {name} already exists")
        self._params[name] = value
        self._param_delayed[name] = dask.delayed(name=name)(lambda: self._params[name])()
        return self._param_delayed[name]

    def create_many(self, d: dict) -> None:
        """Create multiple parameters at once"""
        for k, v in d.items():
            self.create(k, v)

    def get_params(self) -> dict:
        """Get parameters as a dictionary (name -> value)"""
        return {k: v for k, v in self._params.items() if not _kw_is_private(k)}

    def get_delayed(self, name: str) -> Delayed:
        """Get a Delayed object for the chosen parameter"""
        return self._param_delayed[name]

    def update(self, name: str, value: Any) -> None:
        """Permanently update parameter value"""
        if name not in self._params:
            raise KeyError(f"Parameter {name} does not exist")
        self._params[name] = value

    def update_many(self, d: dict) -> None:
        """Permanently update multiple parameter values"""
        for k, v in d.items():
            self.update(k, v)

    @contextmanager
    def context(self, d: dict):
        """Update multiple parameter values within a context"""
        with self._lock_context:
            old_params = dict(**self._params)
            self.update_many(d)
            yield
            self.update_many(old_params)


def delayed_cached(
    name: Optional[str] = None,
    name_prefix: str = "",
    parameters: Optional[dict] = None,
    ignore_args: Optional[bool] = None,
    ignore_kwargs: Optional[Union[bool, List[str]]] = None,
    folder: Union[str, Path] = "cache",
    ftype: str = "pickle",
    kwargs_sep: str = "|",
    nout: Optional[int] = None,
    override: bool = False,
    logger=None,
):
    """Delayed and cache function output on the disk (dask.delayed + dutil.pipeline.cached)

    Parameters: see dutil.pipeline.cached"""

    def decorator(foo):
        """Delay and cache function output on disk (dask.delayed + dutil.pipeline.cached)"""
        # @functools.wraps(foo)  # Can't use: Delayed objects are immutable
        @dask.delayed(name=name, pure=False, nout=nout)
        @cached(
            name=name,
            name_prefix=name_prefix,
            parameters=parameters,
            ignore_args=ignore_args,
            ignore_kwargs=ignore_kwargs,
            folder=folder,
            ftype=ftype,
            kwargs_sep=kwargs_sep,
            nout=nout,
            override=override,
            logger=logger,
        )
        @functools.wraps(foo)
        def new_foo(*args, **kwargs):
            return foo(*args, **kwargs)

        return new_foo

    return decorator


def delayed_compute(tasks, scheduler="threads") -> tuple:
    """Compute values of Delayed objects or load it from cache"""
    results = dask.compute(*tasks, scheduler=scheduler)
    datas = tuple(r.load() if isinstance(r, CachedResultItem) else r for r in results)
    return datas



import Levenshtein  # noqa: F401
from fuzzywuzzy import fuzz

_translation_table = dict.fromkeys(map(ord, r".,`'!@#$()=^"), None)
_replace_dict = {
    " incorporated": " inc",
    " corporation": " corp",
    " company": " co",
    " limited": " ltd",
    " and": " &",
    " corp": " inc",
    " ltd": " inc",
    " llc": " inc",
    " plc": " inc",
}
_abbr_dict = {
    "hp": "hewlett-packard",
}


def _translate(x: str):
    x = x.lower().translate(_translation_table)
    for k, v in _replace_dict.items():
        x = x.replace(k, v)
    if x.startswith("the "):
        x = x[4:] + " the"
    x = x.strip()
    x = _abbr_dict.get(x, x)
    return x


def compare_companies(x: str, y: str) -> float:
    """Compare names of two companies taken into account typical idiosyncrasies in naming

    :param x: first company name
    :param y: second company name
    :return: matching rating from 0 (no match) to 1 (a perfect match)
    """
    x = _translate(x)
    y = _translate(y)
    return fuzz.partial_ratio(x, y) / 100



from __future__ import annotations

import functools
import json
import multiprocessing
import shutil
from pathlib import Path
from typing import Any, Callable, Optional, Union

import dill

# from contextlib import contextmanager
# from contextvars import ContextVar
import numpy as np
import pandas as pd
import pyarrow
import xxhash
from dask.delayed import Delayed
from loguru import logger as _logger

_ = pyarrow.__version__  # set pyarrow dependency explicitly

xxhasher = xxhash.xxh64(seed=42)

MAX_ARG_HASH_LEN = 32  # limit length of hash string
MAX_NAME_LEN = 240  # limit length of cache file name (not counting file extention)


def _hash_ndarray(arr: np.ndarray):
    try:
        data = arr if np.issubdtype(arr.dtype, np.number) else str(arr)
    except TypeError:  # numpy cannot determine data type
        data = str(arr)
    try:
        xxhasher.update(data)
    except ValueError:  # cannot hash M-type array
        data = str(arr)
        xxhasher.update(data)
    h = str(xxhasher.intdigest())
    xxhasher.reset()
    return h


def _hash_obj(obj, max_len: Optional[int] = MAX_ARG_HASH_LEN) -> str:
    if isinstance(obj, np.ndarray):
        h = _hash_ndarray(obj)
    elif isinstance(obj, pd.Series):
        # try:
        #     xxhasher.update(obj.values)
        # except (TypeError, ValueError):
        #     xxhasher.update(str(obj.values))
        h = _hash_ndarray(obj.values)
    elif isinstance(obj, pd.DataFrame):
        h_ = ""
        for c in obj:
            # try:
            #     xxhasher.update(obj[c].values)
            # except (TypeError, ValueError):
            #     xxhasher.update(str(obj[c].values))
            h_ = h_ + _hash_ndarray(obj[c].values)
        xxhasher.update(h_)
        h = str(xxhasher.intdigest())
        xxhasher.reset()
    else:
        h = str(obj)
    if (max_len is not None) and (len(h) > max_len):
        xxhasher.update(h)
        h_sffx = str(xxhasher.intdigest())
        xxhasher.reset()
        h = f"{h[: max_len - len(h_sffx) - 1]}-{h_sffx}"
    return h


def _hash_obj_cached(obj, max_len: int = MAX_ARG_HASH_LEN) -> str:
    if hasattr(obj, "__cached_hash__"):
        h = obj.__cached_hash__()
    else:
        h = _hash_obj(obj, max_len)
    return h


def _kw_is_private(k: str) -> bool:
    return k.startswith("_")


def _get_cache_name(
    name: Optional[str],
    name_prefix: str,
    parameters: Optional[dict],
    ignore_args: Optional[bool],
    ignore_kwargs: Optional[bool],
    kwargs_sep: str,
    foo: Callable,
    args: list,
    kwargs: dict,
    max_name_len: Optional[int] = MAX_NAME_LEN,
) -> str:
    if ignore_args is None:
        ignore_args = parameters is not None
    if ignore_kwargs is None:
        ignore_kwargs = parameters is not None
    if name is None:
        name = foo.__name__
    _n = [name]
    if parameters is not None:
        _n.extend(
            [
                str(k) + kwargs_sep + _hash_obj_cached(v)
                for k, v in parameters.items()
                if not _kw_is_private(k)
            ]
        )
    if not ignore_args:
        _n.extend([_hash_obj_cached(a) for a in args])
    if not ignore_kwargs:
        _n.extend(
            [
                str(k) + kwargs_sep + _hash_obj_cached(v)
                for k, v in kwargs.items()
                if not _kw_is_private(k)
            ]
        )
    elif isinstance(ignore_kwargs, list) or isinstance(ignore_kwargs, set):
        _n.extend(
            [
                str(k) + kwargs_sep + _hash_obj_cached(v)
                for k, v in kwargs.items()
                if k not in ignore_kwargs
            ]
        )
    else:
        assert isinstance(ignore_kwargs, bool)
    full_name = name_prefix + "_".join(_n)
    if (max_name_len is not None) and (len(full_name) > max_name_len):
        xxhasher.update(full_name)
        h_sffx = str(xxhasher.intdigest())
        xxhasher.reset()
        full_name = f"{full_name[: max_name_len - len(h_sffx) - 1]}-{h_sffx}"
    return full_name


def _cached_load(ftype, path):
    if ftype == "parquet":
        data = pd.read_parquet(path)
    elif ftype == "pickle":
        data = dill.load(open(path, "rb"))
    else:
        raise ValueError("ftype {} is not recognized".format(ftype))
    return data


def _cached_save(data, ftype, path) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    if ftype == "parquet":
        data.to_parquet(path, index=False, allow_truncated_timestamps=True)
    elif ftype == "pickle":
        dill.dump(data, open(path, "wb"))
    else:
        raise ValueError("ftype {} is not recognized".format(ftype))


class CacheMeta:
    """Cache meta data (incl. hash)"""

    def __init__(
        self,
        name: str,
        folder: Union[Path, str],
        ftype: str,
        nout: Optional[int],
        hash_value: Optional[str],
    ):
        self.name = name
        self._folder = Path(folder).absolute()
        self.ftype = ftype
        self.nout = nout
        self.hash_value = hash_value

    @staticmethod
    def _get_meta_path(folder: Union[Path, str], name: str) -> Path:
        return Path(folder).absolute() / (name + ".meta")

    @property
    def meta_path(self) -> Path:
        return self._get_meta_path(folder=self._folder, name=self.name)

    @property
    def cache_path(self) -> Union[Path, list[Path]]:
        if self.nout is None:
            return self._folder / (self.name + f".{self.ftype}")
        else:
            return [self._folder / (self.name + f"__{i}.{self.ftype}") for i in range(self.nout)]

    @classmethod
    def from_file(cls, folder: Union[Path, str], name: str) -> CacheMeta:
        meta_path = cls._get_meta_path(folder=folder, name=name)
        if meta_path.exists():
            with open(meta_path, "rt") as f:
                fields = json.load(f)
            # for k, v in fields.items():
            #     if isinstance(v, list):
            #         fields[k] = tuple(v)
            return cls(folder=folder, **fields)
        else:
            raise FileNotFoundError(f"Meta file is missing: {meta_path}")

    def dump_to_file(self) -> None:
        with open(self.meta_path, "wt") as f:
            fields = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
            json.dump(fields, f)


class CachedResult:
    """Lazy loader for cache data"""

    @classmethod
    def from_user(
        cls,
        name,
        name_prefix,
        parameters,
        ignore_args,
        ignore_kwargs,
        folder,
        ftype,
        kwargs_sep,
        foo,
        args,
        kwargs,
        logger,
        nout,
    ):
        cache_name = _get_cache_name(
            name=name,
            name_prefix=name_prefix,
            parameters=parameters,
            ignore_args=ignore_args,
            ignore_kwargs=ignore_kwargs,
            kwargs_sep=kwargs_sep,
            foo=foo,
            args=args,
            kwargs=kwargs,
        )
        try:
            meta = CacheMeta.from_file(folder=folder, name=cache_name)
        except FileNotFoundError as e:
            logger.debug(str(e))
            meta = CacheMeta(
                name=cache_name, folder=folder, ftype=ftype, nout=nout, hash_value=None
            )
        return cls(meta=meta, logger=logger)

    def __init__(self, meta: CacheMeta, logger):
        self.meta = meta
        self.logger = logger
        self._cache_value = None
        self._lock_dump_load = multiprocessing.Lock()
        self._lock_hash = multiprocessing.Lock()

    def load(self) -> Any:
        """Load data from cache"""

        if self._cache_value is None:
            with self._lock_dump_load:
                if self.meta.nout is None:
                    self._cache_value = _cached_load(self.meta.ftype, self.meta.cache_path)
                else:
                    self._cache_value = tuple(
                        _cached_load(self.meta.ftype, cp) for cp in self.meta.cache_path
                    )
            self.logger.debug("Task {}: data has been loaded from cache".format(self.meta.name))
        return self._cache_value

    def dump(self, data):
        """Update and dump data to cache"""

        with self._lock_dump_load:
            self._cache_value = data
            if self.meta.nout is None:
                _cached_save(self._cache_value, self.meta.ftype, self.meta.cache_path)
            else:
                assert len(self._cache_value) == self.meta.nout
                assert len(self.meta.cache_path) == self.meta.nout
                for cv, cp in zip(self._cache_value, self.meta.cache_path):
                    _cached_save(cv, self.meta.ftype, cp)
        self.logger.debug("Task {}: data has been saved to cache".format(self.meta.name))
        self.meta.dump_to_file()

    def __cached_hash__(self):
        """Get hash of cached data

        Used to construct a cache file name
        """

        # Hash may not be required, so it's not automatically computed from data
        if self.meta.hash_value is None:
            with self._lock_hash:
                cache_obj = self.load()  # activates _lock_dump_load
                if self.meta.nout is None:
                    self.meta.hash_value = _hash_obj(cache_obj)
                else:
                    self.meta.hash_value = tuple(_hash_obj(co) for co in cache_obj)
                self.meta.dump_to_file()
            self.logger.debug("Task {}: hash has been computed from data".format(self.meta.name))
        return self.meta.hash_value

    def exists(self):
        return self.meta.meta_path.exists()


class CachedResultItem:
    """Lazy loader for cache data

    A wrapper for CachedResult class to deal with functions that return multiple outputs
    """

    def __init__(self, result: CachedResult, item: Any):
        self.result = result
        self.item = item

    def load(self) -> Any:
        if self.item is None:
            return self.result.load()
        else:
            return self.result.load()[self.item]

    def dump(self, data):
        self.result.dump(data)

    def __cached_hash__(self):
        if self.item is None:
            return self.result.__cached_hash__()
        else:
            return self.result.__cached_hash__()[self.item]


def cached(
    name: Optional[str] = None,
    name_prefix: str = "",
    parameters: Optional[dict] = None,
    ignore_args: Optional[bool] = None,
    ignore_kwargs: Optional[Union[bool, list[str]]] = None,
    folder: Union[str, Path] = "cache",
    ftype: str = "pickle",
    kwargs_sep: str = "|",
    nout: Optional[int] = None,
    override: bool = False,
    logger=None,
):
    """Cache function output on the disk

    Features:
    - Pickle and parquet serialization
    - Special treatment for Delayed objects
    - Lazy cache loading
    - Hashing of complex arguments

    :param name: name of the cache file
        if none, name is constructed from the function name and args
    :param parameters: include these parameters in the name
        only meaningful when `name=None`
        Improtant: parameters starting with _ (underscore) will be ignored
    :param ignore_args: if true, do not add args to the name
    :param ignore_kwargs: if true, do not add kwargs to the name
        it's also possible to specify a list of kwargs to ignore
        Improtant: kwargs starting with _ (underscore) will be ignored
    :param folder: name of the cache folder
    :param ftype: type of the cache file
        'pickle' | 'parquet'
    :param kwargs_sep: string separating a keyword parameter and its value
    :param override: if true, override the existing cache file
    :param logger: if none, use a new logger
    :return: new function
        output is lazily loaded from cache file if it exists, generated otherwise
        .load() to get data
    """

    logger = logger if logger is not None else _logger

    def decorator(foo):
        """Cache function output on the disk"""

        @functools.wraps(foo)
        def new_foo(*args, **kwargs):
            result = CachedResult.from_user(
                name=name,
                name_prefix=name_prefix,
                parameters=parameters,
                ignore_args=ignore_args,
                ignore_kwargs=ignore_kwargs,
                folder=folder,
                ftype=ftype,
                kwargs_sep=kwargs_sep,
                foo=foo,
                args=args,
                kwargs=kwargs,
                logger=logger,
                nout=nout,
            )
            if not override and result.exists():
                # if the result (= cache OR cache + hash) exists, do nothing - just pass it on
                # the cache will be loaded only if required later
                if nout is not None:
                    output = tuple(CachedResultItem(result, i) for i in range(nout))
                else:
                    output = CachedResultItem(result, None)
                logger.info("Task {}: skip (cache exists)".format(result.meta.name))
            else:
                # if the result does not exist, generate data and save cache
                dask_args_detected = any(isinstance(a, Delayed) for a in args)
                dask_kwargs_detected = any(isinstance(v, Delayed) for _, v in kwargs.items())
                if not dask_args_detected and not dask_kwargs_detected:
                    # eager load cache for all arguments
                    args = [a.load() if isinstance(a, CachedResultItem) else a for a in args]
                    kwargs = {
                        k: v.load() if isinstance(v, CachedResultItem) else v
                        for k, v in kwargs.items()
                    }
                    data = foo(*args, **kwargs)
                    result.dump(data)
                    if nout is not None:
                        output = tuple(CachedResultItem(result, i) for i in range(nout))
                    else:
                        output = CachedResultItem(result, None)
                    logger.info(
                        "Task {}: data has been computed and saved to cache".format(
                            result.meta.name
                        )
                    )
                else:
                    # if any of the arguments is a Delayed object, return anything
                    output = foo(*args, **kwargs)
            return output

        return new_foo

    return decorator


def clear_cache(
    folder: Union[str, Path] = "cache",
    ignore_errors: bool = True,
):
    """Clear the cache folder

    :param folder: name of the cache folder
    """
    folder = Path(folder).absolute()
    shutil.rmtree(folder, ignore_errors=ignore_errors)



import pytest

from dutil.string import compare_companies

eps = 0.00001


@pytest.mark.parametrize(
    "name_1, name_2, rating",
    [
        ("Apple Inc.", "Apple Inc.", 1.0),
        ("Apple Inc.", "xxxxx", 0),
        ("Apple Inc.", "Apple Incorporated", 1.0),
        ("Apple Inc.", "Apple Limited", 1.0),
        ("Apple Inc.", "Apple", 1.0),
        ("Aarons Holdings Company Inc.", "Aaron's, Inc.", 0.7),
        ("The Tea Company", "Tea Company (the)", 1.0),
        ("HP", "Hewlett-Packard", 1.0),
        ("HP", "HewlettPackard", 0.93),
    ],
)
def test_compare_companies(name_1: str, name_2: str, rating: float):
    computed_rating = compare_companies(name_1, name_2)
    assert abs(rating - computed_rating) < eps



"""
Data persistance and pipelining tools
"""

from dutil.pipeline._cached import CachedResultItem, cached, clear_cache  # noqa: F401
from dutil.pipeline._dask import (  # noqa: F401
    DelayedParameter,
    DelayedParameters,
    delayed_cached,
    delayed_compute,
)



import numpy as np
import pandas as pd
import pytest

from dutil.stats import mean_lower, mean_upper


@pytest.mark.parametrize(
    "data, expected",
    [
        (np.array([0, 1, 5, -1]), 3.0),
        (pd.Series([0, 1, 5, -1]), 3.0),
    ],
)
def test_mean_upper_assert_equal(data, expected):
    actual = mean_upper(data)
    assert actual == expected


@pytest.mark.parametrize(
    "data, expected",
    [
        (np.array([0, 1, 5, -1]), -0.5),
        (pd.Series([0, 1, 5, -1]), -0.5),
    ],
)
def test_mean_lower_assert_equal(data, expected):
    actual = mean_lower(data)
    assert actual == expected



import datetime as dt
import time
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from dask import delayed

from dutil.pipeline import (
    DelayedParameter,
    DelayedParameters,
    cached,
    clear_cache,
    delayed_cached,
    delayed_compute,
)

CACHE_DIR = Path("cache/temp/")
EPS = 0.00001


def _count_cache_files() -> int:
    return len([f for f in CACHE_DIR.iterdir() if f.is_file()])


@pytest.mark.parametrize(
    "data, ftype",
    [
        ((0, 1, 3, 5, -1), "pickle"),
        ((0, 1.0, 3232.22, 5.0, -1.0, None), "pickle"),
        ([0, 1, 3, 5, -1], "pickle"),
        ([0, 1.0, 3232.22, 5.0, -1.0, None], "pickle"),
        (np.array([0, 1.0, 3232.22, 5.0, -1.0]), "pickle"),
        (np.array([dt.datetime(2019, 1, 1)] * 3), "pickle"),
        (pd.Series([0, 1, 3, 5, -1]), "pickle"),
        (pd.Series([0, 1.0, 3232.22, 5.0, -1.0, np.nan]), "pickle"),
        (pd.Series([1, 2, 3, 4], dtype="category"), "pickle"),
        (pd.Series(pd.date_range("2018-01-01", periods=5)), "pickle"),
        (
            pd.DataFrame(
                {
                    "a": [0, 1, 3, 5, -1],
                    "b": [2, 1, 0, 0, 14],
                }
            ),
            "pickle",
        ),
        (
            pd.DataFrame(
                {
                    "a": [0, 1.0, 3232.22, -1.0, np.nan],
                    "b": ["a", "b", "c", "ee", "14"],
                    "c": [
                        dt.datetime(2018, 1, 1),
                        dt.datetime(2019, 1, 1),
                        dt.datetime(2020, 1, 1),
                        dt.datetime(2021, 1, 1),
                        dt.datetime(2022, 1, 1),
                    ],
                }
            ),
            "pickle",
        ),
        (
            pd.DataFrame(
                {
                    "a": [0, 1, 3, 5, -1],
                    "b": [2, 1, 0, 0, 14],
                }
            ),
            "parquet",
        ),
        (
            pd.DataFrame(
                {
                    "a": [0, 1.0, 3232.22, -1.0, np.nan],
                    "b": ["a", "b", "c", "ee", "14"],
                    "c": [
                        dt.datetime(2018, 1, 1),
                        dt.datetime(2019, 1, 1),
                        dt.datetime(2020, 1, 1),
                        dt.datetime(2021, 1, 1),
                        dt.datetime(2022, 1, 1),
                    ],
                    "d": [
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                    ],
                }
            ),
            "parquet",
        ),
        (
            pd.DataFrame(
                {
                    "a": [0, 1.0, 3232.22, -1.0, np.nan],
                    "b": [
                        dt.timedelta(hours=1),
                        dt.timedelta(hours=1),
                        dt.timedelta(hours=1),
                        dt.timedelta(hours=1),
                        dt.timedelta(hours=1),
                    ],
                }
            ),
            "pickle",
        ),
    ],
)
def test_cached_load_and_hash(data, ftype):
    @cached(folder=CACHE_DIR, ftype=ftype, override=False)
    def load_data():
        return data

    @cached(folder=CACHE_DIR, ftype="pickle", override=False)
    def compute_data(data):
        return 0

    clear_cache(CACHE_DIR)
    loaded = load_data().load()
    _ = compute_data(loaded).load()
    assert _count_cache_files() == 4
    loaded = load_data().load()
    _ = compute_data(loaded).load()
    assert _count_cache_files() == 4

    if isinstance(data, pd.Series):
        pd.testing.assert_series_equal(loaded, data)
    elif isinstance(data, pd.DataFrame):
        pd.testing.assert_frame_equal(loaded, data)
    elif isinstance(data, np.ndarray):
        np.testing.assert_equal(loaded, data)
    else:
        assert loaded == data


@pytest.mark.parametrize(
    "data, ftype, eps, ts",
    [
        (
            pd.DataFrame(
                {
                    "a": [0, 1, 3, 5, -1],
                    "b": [2, 1, 0, 0, 14],
                }
            ),
            "parquet",
            0.1,
            pd.Timestamp("2018-01-01"),
        ),
        (
            pd.DataFrame(
                {
                    "a": [0, 1.0, 3232.22, -1.0, np.nan],
                    "b": ["a", "b", "c", "ee", "14"],
                    "c": [
                        dt.datetime(2018, 1, 1),
                        dt.datetime(2019, 1, 1),
                        dt.datetime(2020, 1, 1),
                        dt.datetime(2021, 1, 1),
                        dt.datetime(2022, 1, 1),
                    ],
                    "d": [
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                    ],
                }
            ),
            "parquet",
            0.1,
            pd.Timestamp("2018-01-01"),
        ),
        (
            pd.DataFrame(
                {
                    "a": [0, 1.0, 3232.22, -1.0, np.nan],
                    "b": [
                        dt.timedelta(hours=1),
                        dt.timedelta(hours=1),
                        dt.timedelta(hours=1),
                        dt.timedelta(hours=1),
                        dt.timedelta(hours=1),
                    ],
                }
            ),
            "pickle",
            0.1,
            pd.Timestamp("2018-01-01"),
        ),
    ],
)
def test_cached_with_args_kwargs_load(data, ftype, eps, ts):
    @cached(folder=CACHE_DIR, ftype=ftype, override=False)
    def load_data(eps, ts):
        assert eps > 0
        assert ts > pd.Timestamp("2000-01-01")
        return data

    clear_cache(CACHE_DIR)
    _ = load_data(eps, ts=ts).load()
    assert _count_cache_files() == 2
    loaded = load_data(eps, ts=ts).load()
    assert _count_cache_files() == 2

    if isinstance(data, pd.Series):
        pd.testing.assert_series_equal(loaded, data)
    elif isinstance(data, pd.DataFrame):
        pd.testing.assert_frame_equal(loaded, data)
    elif isinstance(data, np.ndarray):
        np.testing.assert_equal(loaded, data)
    else:
        assert loaded == data, str(loaded)


@pytest.mark.parametrize(
    "data, output, ftype",
    [
        (
            pd.DataFrame(
                {
                    "a": [0, 1, 3, 5, -1],
                    "b": [2, 1, 0, 0, 14],
                }
            ),
            pd.DataFrame(
                {
                    "a": [0, 1, 3, 5, -1],
                    "b": [2, 1, 0, 0, 14],
                }
            ),
            "parquet",
        ),
        (
            pd.DataFrame(
                {
                    "a": [0.5, np.nan, np.nan],
                    "b": ["a", "b", "14"],
                    "c": [
                        dt.datetime(2018, 1, 1),
                        dt.datetime(2019, 1, 1),
                        dt.datetime(2022, 1, 1),
                    ],
                    "d": [
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                        pd.Timestamp("2018-01-01"),
                    ],
                }
            ),
            pd.DataFrame(
                {
                    "a": [0.5],
                    "b": ["a"],
                    "c": [dt.datetime(2018, 1, 1)],
                    "d": [pd.Timestamp("2018-01-01")],
                }
            ),
            "parquet",
        ),
        (
            pd.DataFrame(
                {
                    "a": [0, 1.0, np.nan],
                    "b": [dt.timedelta(hours=1), dt.timedelta(hours=1), dt.timedelta(hours=1)],
                }
            ),
            pd.DataFrame(
                {
                    "a": [0, 1.0],
                    "b": [dt.timedelta(hours=1), dt.timedelta(hours=1)],
                }
            ),
            "pickle",
        ),
    ],
)
def test_cached_with_chained_df_load_and_hash(data, output, ftype):
    @cached(folder=CACHE_DIR, ftype=ftype, override=False)
    def load_data():
        return data

    @cached(folder=CACHE_DIR, ftype=ftype, override=False)
    def process_data(df):
        return df.dropna()

    clear_cache(CACHE_DIR)
    df = load_data()
    _ = process_data(df).load()

    df = load_data()
    processed = process_data(df).load()

    if isinstance(data, pd.Series):
        pd.testing.assert_series_equal(processed, output)
    elif isinstance(data, pd.DataFrame):
        pd.testing.assert_frame_equal(processed, output)
    elif isinstance(data, np.ndarray):
        np.testing.assert_equal(processed, output)
    else:
        assert processed == output


@pytest.mark.parametrize(
    "data, ftype, eps, ts",
    [
        (
            pd.DataFrame(
                {
                    "a": [0, 1, 3, 5, -1],
                    "b": [2, 1, 0, 0, 14],
                }
            ),
            "parquet",
            0.1,
            pd.Timestamp("2018-01-01"),
        ),
    ],
)
def test_dask_cached_with_args_kwargs_load_compute(data, ftype, eps, ts):
    @delayed()
    @cached(folder=CACHE_DIR, ftype=ftype, override=False)
    def load_data(eps, ts):
        assert eps > 0
        assert ts > pd.Timestamp("2000-01-01")
        return data

    clear_cache(CACHE_DIR)
    r = load_data(eps, ts=ts)
    _ = r.compute()
    loaded = r.compute().load()

    if isinstance(data, pd.Series):
        pd.testing.assert_series_equal(loaded, data)
    elif isinstance(data, pd.DataFrame):
        pd.testing.assert_frame_equal(loaded, data)
    elif isinstance(data, np.ndarray):
        np.testing.assert_equal(loaded, data)
    else:
        assert loaded == data


def test_cached_with_args_kwargs_partial_ignore():
    @cached(folder=CACHE_DIR, ignore_kwargs=["ts"])
    def load_data(eps, ts):
        time.sleep(1.0)
        assert eps > 0
        assert ts > pd.Timestamp("2000-01-01")
        return eps

    clear_cache(CACHE_DIR)
    start = dt.datetime.utcnow()
    res1 = load_data(eps=0.1, ts=pd.Timestamp("2010-01-01")).load()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay > 0.95

    start = dt.datetime.utcnow()
    res2 = load_data(eps=0.1, ts=pd.Timestamp("2012-01-01")).load()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert res1 == res2

    start = dt.datetime.utcnow()
    res3 = load_data(eps=0.2, ts=pd.Timestamp("2012-01-01")).load()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay > 0.95
    assert res1 != res3


def test_cached_load_time():
    @cached(folder=CACHE_DIR, override=False)
    def load_data():
        time.sleep(1)
        return 1

    clear_cache(CACHE_DIR)

    start = dt.datetime.utcnow()
    _ = load_data().load()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay > 0.95

    start = dt.datetime.utcnow()
    _ = load_data().load()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95


def test_cached_long_name_one_long_argument():
    @cached(folder=CACHE_DIR, override=False)
    def compute(line):
        time.sleep(1)
        return len(line)

    clear_cache(CACHE_DIR)

    start = dt.datetime.utcnow()
    _ = compute("a" * 300).load()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay > 0.95

    start = dt.datetime.utcnow()
    _ = compute("a" * 300).load()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95


def test_cached_long_name_many_arguments():
    @cached(folder=CACHE_DIR, override=False)
    def compute(*args):
        time.sleep(1)
        return sum(len(line) for line in args)

    clear_cache(CACHE_DIR)

    start = dt.datetime.utcnow()
    _ = compute(*("a" * 20 for _ in range(20))).load()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay > 0.95

    start = dt.datetime.utcnow()
    _ = compute(*("a" * 20 for _ in range(20))).load()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95


def test_dask_cached_load_time():
    @delayed()
    @cached(folder=CACHE_DIR, override=False)
    def load_data():
        time.sleep(1)
        return 1

    clear_cache(CACHE_DIR)

    start = dt.datetime.utcnow()
    r = load_data()
    r.compute()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay > 0.95

    start = dt.datetime.utcnow()
    r = load_data()
    r.compute()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95


def test_dask_pipeline():
    clear_cache(CACHE_DIR)

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_1():
        time.sleep(1)
        return 5

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_2():
        time.sleep(1)
        return 3

    @delayed()
    @cached(folder=CACHE_DIR)
    def add(x, y):
        return x + y

    d1 = load_data_1()
    d2 = load_data_2()
    r = add(d1, d2)

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95
    assert output == 8

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert output == 8


def test_dask_pipeline_sequential_runs():
    clear_cache(CACHE_DIR)

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_1():
        time.sleep(1)
        return 5

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_2():
        time.sleep(1)
        return 3

    @delayed()
    @cached(folder=CACHE_DIR)
    def add(x, y):
        return x + y

    d1 = load_data_1()
    d2 = load_data_2()
    r = add(d1, d2)

    start = dt.datetime.utcnow()
    d1_, d2_ = delayed_compute((d1, d2))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert output == 8


def test_dask_pipeline_with_parameters():
    clear_cache(CACHE_DIR)

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_1(ts: dt.datetime):
        assert ts > dt.datetime(2019, 1, 1)
        time.sleep(1)
        return 5

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_2(fix: float):
        time.sleep(1)
        return 3 + fix

    @delayed()
    @cached(folder=CACHE_DIR)
    def add(x, y):
        return x + y

    params = DelayedParameters()
    ts = params.create("ts", value=dt.datetime(2020, 1, 1))
    fix = params.create("fix", value=0.5)
    d1 = load_data_1(ts)
    d2 = load_data_2(fix)
    r = add(d1, d2)

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95
    assert abs(output - 8.5) < EPS

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert abs(output - 8.5) < EPS

    params.update_many({"ts": dt.datetime(2020, 2, 1), "fix": 1.5})
    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95
    assert abs(output - 9.5) < EPS

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert abs(output - 9.5) < EPS


def test_dask_pipeline_with_parameters_create_many():
    clear_cache(CACHE_DIR)

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_1(ts: dt.datetime):
        assert ts > dt.datetime(2019, 1, 1)
        time.sleep(1)
        return 5

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_2(fix: float):
        time.sleep(1)
        return 3 + fix

    @delayed()
    @cached(folder=CACHE_DIR)
    def add(x, y):
        return x + y

    params = DelayedParameters()
    params.create_many(
        {
            "ts": dt.datetime(2020, 1, 1),
            "fix": 0.5,
        }
    )
    print(params.get_params())
    d2 = load_data_2(params.get_delayed("fix"))
    d1 = load_data_1(params.get_delayed("ts"))
    r = add(d1, d2)

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95
    assert abs(output - 8.5) < EPS


def test_dask_pipeline_with_parameters_context():
    clear_cache(CACHE_DIR)

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_1(ts: dt.datetime):
        assert ts > dt.datetime(2019, 1, 1)
        return 5

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_2(fix: float):
        return 3 + fix

    @delayed()
    @cached(folder=CACHE_DIR)
    def add(x, y):
        return x + y

    params = DelayedParameters()
    ts = params.create("ts", value=dt.datetime(2020, 1, 1))
    fix = params.create("fix", value=0.5)
    d1 = load_data_1(ts)
    d2 = load_data_2(fix)
    r = add(d1, d2)

    (output,) = delayed_compute((r,))
    assert abs(output - 8.5) < EPS

    with params.context({"ts": dt.datetime(2020, 2, 1), "fix": 1.5}):
        (output,) = delayed_compute((r,))
        assert abs(output - 9.5) < EPS

    (output,) = delayed_compute((r,))
    assert abs(output - 8.5) < EPS


def test_dask_pipeline_with_parameters_private():
    clear_cache(CACHE_DIR)

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_1(_ts: dt.datetime):
        time.sleep(1)
        assert _ts > dt.datetime(2019, 1, 1)
        return 5

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_2(fix: float):
        time.sleep(1)
        return 3 + fix

    @delayed()
    @cached(folder=CACHE_DIR)
    def add(x, y):
        return x + y

    params = DelayedParameters()
    _ts = params.create("_ts", value=dt.datetime(2020, 1, 1))
    fix = params.create("fix", value=0.5)
    d1 = load_data_1(_ts=_ts)
    d2 = load_data_2(fix=fix)
    r = add(d1, d2)

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95

    with params.context({"_ts": dt.datetime(2020, 2, 1)}):
        start = dt.datetime.utcnow()
        (output,) = delayed_compute((r,))
        delay = (dt.datetime.utcnow() - start).total_seconds()
        assert delay < 0.95


def test_dask_pipeline_with_parameters_2():
    clear_cache(CACHE_DIR)

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_1(ts: dt.datetime):
        assert ts > dt.datetime(2019, 1, 1)
        time.sleep(1)
        return 5

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_2(fix: float):
        time.sleep(1)
        return 3 + fix

    @delayed()
    @cached(folder=CACHE_DIR)
    def add(x, y):
        return x + y

    ts = DelayedParameter("ts", value=dt.datetime(2020, 1, 1))
    fix = DelayedParameter("fix", value=0.5)
    d1 = load_data_1(ts())
    d2 = load_data_2(fix())
    r = add(d1, d2)

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95
    assert abs(output - 8.5) < EPS

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert abs(output - 8.5) < EPS

    ts.set(dt.datetime(2020, 2, 1))
    fix.set(1.5)
    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95
    assert abs(output - 9.5) < EPS

    start = dt.datetime.utcnow()
    (output,) = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert abs(output - 9.5) < EPS


def test_dask_pipeline_with_parameters_2_context():
    clear_cache(CACHE_DIR)

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_1(ts: dt.datetime):
        assert ts > dt.datetime(2019, 1, 1)
        return 5

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data_2(fix: float):
        return 3 + fix

    @delayed()
    @cached(folder=CACHE_DIR)
    def add(x, y):
        return x + y

    ts = DelayedParameter("ts", value=dt.datetime(2020, 1, 1))
    fix = DelayedParameter("fix", value=0.5)
    d1 = load_data_1(ts())
    d2 = load_data_2(fix())
    r = add(d1, d2)

    (output,) = delayed_compute((r,))
    assert abs(output - 8.5) < EPS

    with ts.context(dt.datetime(2020, 2, 1)), fix.context(1.5):
        (output,) = delayed_compute((r,))
        assert abs(output - 9.5) < EPS

    (output,) = delayed_compute((r,))
    assert abs(output - 8.5) < EPS


def test_dask_pipeline_multiple_outputs():
    clear_cache(CACHE_DIR)

    @delayed()
    @cached(folder=CACHE_DIR)
    def load_data():
        # time.sleep(1)
        return [1, 1, 1, 2, 2, 2]

    @delayed(nout=2)
    @cached(folder=CACHE_DIR, nout=2)
    def split_data(data):
        return data[:3], data[3:]

    @delayed()
    @cached(folder=CACHE_DIR)
    def compute_sum(arr):
        # time.sleep(1)
        return sum(arr)

    data = load_data()
    x, y = split_data(data)
    xsum = compute_sum(x)
    ysum = compute_sum(y)

    # start = dt.datetime.utcnow()
    (x_, y_, xsum_, ysum_) = delayed_compute((x, y, xsum, ysum))
    # delay = (dt.datetime.utcnow() - start).total_seconds()
    # assert 0.95 < delay < 1.95
    assert x_ == [1, 1, 1]
    assert y_ == [2, 2, 2]
    assert xsum_ == 3
    assert ysum_ == 6


def test_delayed_cached_load_time():
    @delayed_cached(folder=CACHE_DIR, override=False)
    def load_data():
        time.sleep(1)
        return 1

    clear_cache(CACHE_DIR)

    start = dt.datetime.utcnow()
    r = load_data()
    _ = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay > 0.95

    start = dt.datetime.utcnow()
    r = load_data()
    _ = delayed_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95


def test_delayed_cached_another_cache_dir():
    another_cache_dir = "cache/temp2/"

    @delayed_cached(folder=another_cache_dir)
    def load_data():
        return 1

    clear_cache(another_cache_dir)
    r = load_data()
    _ = delayed_compute((r,))

    assert (Path(another_cache_dir) / "load_data.pickle").exists()



import os
import pathlib

from setuptools import find_packages, setup

PATH = pathlib.Path(__file__).parent
VERSION = os.getenv("VERSION", "dev")
LONG_DESCRIPTION = (PATH / "README.md").read_text()

setup(
    name="dutil",
    version=VERSION,
    description="A few useful tools for data wrangling",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Yaroslav Kopotilov",
    license="Apache License, Version 2.0",
    author_email="UNKNOWN",
    url="https://github.com/mysterious-ben/dutil",
    install_requires=[
        "numpy",
        "pandas",
        "pyarrow",
        "dill",
        "loguru",
        "xxhash",
        "dask[delayed]",
        "ipython",
        "fuzzywuzzy",
        "python_Levenshtein",
    ],
    entry_points={"console_scripts": []},
    packages=find_packages(),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.7",
    test_suite="tests",
    setup_requires=[
        "pytest-runner",
        "setuptools",
        "wheel",
    ],
    extras_require={
        "dev": [
            "pre-commit",
        ]
    },
    tests_require=[
        "pytest",
    ],
)


# --- pip upload ---
# python setup.py sdist bdist_wheel
# twine check dist/*
# twine upload dist/*
